# coding: utf-8

"""
This module provides a Renderer class to render templates.

"""

import cgi
import sys

from .context import Context
from .loader import Loader
from .renderengine import RenderEngine


markupsafe = None
try:
    import markupsafe
except ImportError:
    pass


class Renderer(object):

    """
    A class for rendering mustache templates.

    This class supports several rendering options which are described in
    the constructor's docstring.  Among these, the constructor supports
    passing a custom template loader.

    Here is an example of passing a custom template loader to render a
    template using partials loaded from a string-string dictionary.

    >>> partials = {'partial': 'Hello, {{thing}}!'}
    >>> renderer = Renderer(loader=partials)
    >>> renderer.render('{{>partial}}', {'thing': 'world'})
    u'Hello, world!'

    """

    def __init__(self, loader=None, default_encoding=None, decode_errors='strict',
                 output_encoding=None, escape=None):
        """
        Construct an instance.

        Arguments:

          loader: the object (e.g. pystache.Loader or dictionary) that will
            load templates during the rendering process, for example when
            loading a partial.
                The loader should have a get() method that accepts a string
            and returns the corresponding template as a string, preferably
            as a unicode string.  If there is no template with that name,
            the method should either return None (as dict.get() does) or
            raise an exception.
                Defaults to constructing a Loader instance with
            default_encoding and decode_errors passed as the encoding and
            decode_errors arguments, respectively.

          output_encoding: the encoding to use when rendering to a string.
            The argument should be the name of an encoding as a string, for
            example "utf-8".  See the render() method's documentation for
            more information.

          escape: the function used to escape variable tag values when
            rendering a template.  The function should accept a unicode
            string (or subclass of unicode) and return an escaped string
            that is again unicode (or a subclass of unicode).
                This function need not handle strings of type `str` because
            this class will only pass it unicode strings.  The constructor
            assigns this function to the constructed instance's escape()
            method.
                The argument defaults to markupsafe.escape when markupsafe
            is importable and cgi.escape otherwise.  To disable escaping
            entirely, one can pass `lambda u: u` as the escape function,
            for example.

          default_encoding: the name of the encoding to use when converting
            to unicode any strings of type `str` encountered during the
            rendering process.  The name will be passed as the "encoding"
            argument to the built-in function unicode().  Defaults to the
            encoding name returned by sys.getdefaultencoding().

          decode_errors: the string to pass as the "errors" argument to the
            built-in function unicode() when converting to unicode any
            strings of type `str` encountered during the rendering process.
            Defaults to "strict".

        """
        if default_encoding is None:
            default_encoding = sys.getdefaultencoding()

        if escape is None:
            # TODO: use 'quote=True' with cgi.escape and add tests.
            escape = markupsafe.escape if markupsafe else cgi.escape

        if loader is None:
            loader = Loader(encoding=default_encoding, decode_errors=decode_errors)

        self._literal = markupsafe.Markup if markupsafe else unicode

        self.decode_errors = decode_errors
        self.default_encoding = default_encoding
        self.escape = escape
        self.loader = loader
        self.output_encoding = output_encoding

    def _to_unicode_soft(self, s):
        """
        Convert an str or unicode string to a unicode string (or subclass).

        """
        # Avoid the "double-decoding" TypeError.
        return s if isinstance(s, unicode) else self.unicode(s)

    def _to_unicode_hard(self, s):
        """
        Convert an str or unicode string to a unicode string (not subclass).

        """
        return unicode(self._to_unicode_soft(s))

    def _escape_to_unicode(self, s):
        """
        Convert an str or unicode string to unicode, and escape it.

        Returns a unicode string (not subclass).

        """
        return unicode(self.escape(self._to_unicode_soft(s)))

    def unicode(self, s):
        """
        Convert a string to unicode, using default_encoding and decode_errors.

        Raises:

          TypeError: Because this method calls Python's built-in unicode()
            function, this method raises the following exception if the
            given string is already unicode:

              TypeError: decoding Unicode is not supported

        """
        # TODO: Wrap UnicodeDecodeErrors with a message about setting
        # the default_encoding and decode_errors attributes.
        return unicode(s, self.default_encoding, self.decode_errors)

    def _make_context(self, context, **kwargs):
        """
        Initialize the context attribute.

        """
        if context is None:
            context = {}

        if isinstance(context, Context):
            context = context.copy()
        else:
            context = Context(context)

        if kwargs:
            context.push(kwargs)

        return context

    def _make_load_partial(self):
        def load_partial(name):
            template = self.loader.get(name)

            if template is None:
                # TODO: make a TemplateNotFoundException type that provides
                # the original loader as an attribute.
                raise Exception("Partial not found with name: %s" % repr(name))

            # RenderEngine requires that the return value be unicode.
            return self._to_unicode_hard(template)

        return load_partial

    def _make_render_engine(self):
        """
        Return a RenderEngine instance for rendering.

        """
        load_partial = self._make_load_partial()

        engine = RenderEngine(load_partial=load_partial,
                              literal=self._to_unicode_hard,
                              escape=self._escape_to_unicode)
        return engine

    def render(self, template, context=None, **kwargs):
        """
        Render the given template using the given context.

        Returns:

          If the output_encoding attribute is None, the return value is
          markupsafe.Markup if markup was importable and unicode if not.
          Otherwise, the return value is encoded to a string of type str
          using the output encoding named by the output_encoding attribute.

        Arguments:

          template: a template string that is either unicode or of type str.
            If the string has type str, it is first converted to unicode
            using this instance's default_encoding and decode_errors
            attributes.  See the constructor docstring for more information.

          context: a dictionary, Context, or object (e.g. a View instance).

          **kwargs: additional key values to add to the context when
            rendering.  These values take precedence over the context on
            any key conflicts.

        """
        engine = self._make_render_engine()
        context = self._make_context(context, **kwargs)

        # RenderEngine.render() requires that the template string be unicode.
        template = self._to_unicode_hard(template)

        rendered = engine.render(template, context)
        rendered = self._literal(rendered)

        if self.output_encoding is not None:
            rendered = rendered.encode(self.output_encoding)

        return rendered
